@@ -60,20 +60,7 @@
  * them as to usage or instruct them to close them.
  */
 
-size_t open_fd_count = 0;
-
-/* XXX dang locking
- * - FD locking (open, close, is_open) - was content lock
- *   will be fixed in conversion to support_ex
- */
-
-
-bool fsal_is_open(struct fsal_obj_handle *obj)
-{
-	if ((obj == NULL) || (obj->type != REGULAR_FILE))
-		return false;
-	return (obj->obj_ops.status(obj) != FSAL_O_CLOSED);
-}
+size_t open_fd_count;
 
 static bool fsal_not_in_group_list(gid_t gid)
 {
@@ -130,7 +117,7 @@ static fsal_status_t check_open_permission(struct fsal_obj_handle *obj,
 	 * verifier later, so this allows a replay of
 	 * open("foo", O_RDWR | O_CREAT | O_EXCL, 0) to succeed).
 	 */
-	status = obj->obj_ops.test_access(obj, access_mask,
+	status = obj->obj_ops->test_access(obj, access_mask,
 					  NULL, NULL, exclusive_create);
 
 	if (!FSAL_IS_ERROR(status)) {
@@ -202,11 +189,11 @@ static fsal_status_t fsal_check_setattr_perms(struct fsal_obj_handle *obj,
 	}
 
 	fsal_prepare_attrs(current,
-			   op_ctx->fsal_export->
-				exp_ops.fs_supported_attrs(op_ctx->fsal_export)
+			   op_ctx->fsal_export->exp_ops.fs_supported_attrs(
+							op_ctx->fsal_export)
 			   & (ATTRS_CREDS | ATTR_MODE | ATTR_ACL));
 
-	status = obj->obj_ops.getattrs(obj, current);
+	status = obj->obj_ops->getattrs(obj, current);
 
 	if (FSAL_IS_ERROR(status))
 		return status;
@@ -338,7 +325,7 @@ static fsal_status_t fsal_check_setattr_perms(struct fsal_obj_handle *obj,
 	}
 
 	if (current->acl) {
-		status = obj->obj_ops.test_access(obj, access_check, NULL,
+		status = obj->obj_ops->test_access(obj, access_check, NULL,
 						  NULL, false);
 		note = " (checked ACL)";
 		goto out;
@@ -351,7 +338,7 @@ static fsal_status_t fsal_check_setattr_perms(struct fsal_obj_handle *obj,
 		goto out;
 	}
 
-	status = obj->obj_ops.test_access(obj, FSAL_W_OK, NULL, NULL, false);
+	status = obj->obj_ops->test_access(obj, FSAL_W_OK, NULL, NULL, false);
 
 	note = " (checked mode)";
 
@@ -403,7 +390,7 @@ fsal_status_t open2_by_name(struct fsal_obj_handle *in_obj,
 	if (FSAL_IS_ERROR(status))
 		return status;
 
-	status = in_obj->obj_ops.open2(in_obj,
+	status = in_obj->obj_ops->open2(in_obj,
 				       state,
 				       openflags,
 				       createmode,
@@ -445,7 +432,7 @@ fsal_status_t open2_by_name(struct fsal_obj_handle *in_obj,
 		 reason, fsal_err_txt(status));
 
 	if (state != NULL)
-		close_status = (*obj)->obj_ops.close2(*obj, state);
+		close_status = (*obj)->obj_ops->close2(*obj, state);
 	else
 		close_status = fsal_close(*obj);
 
@@ -567,24 +554,13 @@ fsal_status_t fsal_setattr(struct fsal_obj_handle *obj, bool bypass,
 		attr->mode &= ~S_ISGID;
 	}
 
-	if (obj->fsal->m_ops.support_ex(obj)) {
-		status = obj->obj_ops.setattr2(obj, bypass, state, attr);
-		if (FSAL_IS_ERROR(status)) {
-			if (status.major == ERR_FSAL_STALE) {
-				LogEvent(COMPONENT_FSAL,
-					 "FSAL returned STALE from setattr2");
-			}
-			return status;
-		}
-	} else {
-		status = obj->obj_ops.setattrs(obj, attr);
-		if (FSAL_IS_ERROR(status)) {
-			if (status.major == ERR_FSAL_STALE) {
-				LogEvent(COMPONENT_FSAL,
-					 "FSAL returned STALE from setattrs");
-			}
-			return status;
+	status = obj->obj_ops->setattr2(obj, bypass, state, attr);
+	if (FSAL_IS_ERROR(status)) {
+		if (status.major == ERR_FSAL_STALE) {
+			LogEvent(COMPONENT_FSAL,
+				 "FSAL returned STALE from setattr2");
 		}
+		return status;
 	}
 
 	if (!is_superuser)  {
@@ -609,7 +585,7 @@ fsal_status_t fsal_readlink(struct fsal_obj_handle *obj,
 		return fsalstat(ERR_FSAL_BADTYPE, 0);
 
 	/* Never refresh.  FSAL_MDCACHE will override for cached FSALs. */
-	return obj->obj_ops.readlink(obj, link_content, false);
+	return obj->obj_ops->readlink(obj, link_content, false);
 }
 
 /**
@@ -659,7 +635,7 @@ fsal_status_t fsal_link(struct fsal_obj_handle *obj,
 
 	/* Rather than performing a lookup first, just try to make the
 	   link and return the FSAL's error if it fails. */
-	status = obj->obj_ops.link(obj, dest_dir, name);
+	status = obj->obj_ops->link(obj, dest_dir, name);
 	return status;
 }
 
@@ -697,14 +673,14 @@ fsal_status_t fsal_lookup(struct fsal_obj_handle *parent,
 		return fsal_status;
 
 	if (strcmp(name, ".") == 0) {
-		parent->obj_ops.get_ref(parent);
+		parent->obj_ops->get_ref(parent);
 		*obj = parent;
-		return fsalstat(ERR_FSAL_NO_ERROR, 0);
+		return get_optional_attrs(*obj, attrs_out);
 	} else if (strcmp(name, "..") == 0)
 		return fsal_lookupp(parent, obj, attrs_out);
 
 
-	return parent->obj_ops.lookup(parent, name, obj, attrs_out);
+	return parent->obj_ops->lookup(parent, name, obj, attrs_out);
 }
 
 /**
@@ -727,7 +703,9 @@ fsal_status_t fsal_lookupp(struct fsal_obj_handle *obj,
 		fsal_status_t status = {0, 0};
 		struct fsal_obj_handle *root_obj = NULL;
 
-		status = nfs_export_get_root_entry(op_ctx->ctx_export, &root_obj);
+		status = nfs_export_get_root_entry(op_ctx->ctx_export,
+						   &root_obj);
+
 		if (FSAL_IS_ERROR(status))
 			return status;
 
@@ -742,18 +720,18 @@ fsal_status_t fsal_lookupp(struct fsal_obj_handle *obj,
 				/* Need to return the attributes of the
 				 * current object.
 				 */
-				return obj->obj_ops.getattrs(obj, attrs_out);
+				return obj->obj_ops->getattrs(obj, attrs_out);
 			} else {
 				/* Success */
 				return fsalstat(ERR_FSAL_NO_ERROR, 0);
 			}
 		} else {
 			/* Return entry from nfs_export_get_root_entry */
-			root_obj->obj_ops.put_ref(root_obj);
+			root_obj->obj_ops->put_ref(root_obj);
 		}
 	}
 
-	return obj->obj_ops.lookup(obj, "..", parent, attrs_out);
+	return obj->obj_ops->lookup(obj, "..", parent, attrs_out);
 }
 
 /**
@@ -815,9 +793,6 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 {
 	fsal_status_t status = { 0, 0 };
 	attrmask_t orig_mask = attrs->valid_mask;
-	uint64_t owner = attrs->owner;
-	uint64_t group = attrs->group;
-	bool support_ex = parent->fsal->m_ops.support_ex(parent);
 
 	if ((type != REGULAR_FILE) && (type != DIRECTORY)
 	    && (type != SYMBOLIC_LINK) && (type != SOCKET_FILE)
@@ -831,22 +806,16 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 		goto out;
 	}
 
-	if (!support_ex) {
-		/* For old API, make sure owner/group attr matches op_ctx */
-		attrs->valid_mask |= ATTR_OWNER | ATTR_GROUP;
-		attrs->owner = op_ctx->creds->caller_uid;
-		attrs->group = op_ctx->creds->caller_gid;
-	} else {
-		/* For support_ex API, turn off owner and/or group attr
-		 * if they are the same as the credentials.
-		 */
-		if ((attrs->valid_mask & ATTR_OWNER) &&
-		    attrs->owner == op_ctx->creds->caller_uid)
-			FSAL_UNSET_MASK(attrs->valid_mask, ATTR_OWNER);
-		if ((attrs->valid_mask & ATTR_GROUP) &&
-		    attrs->group == op_ctx->creds->caller_gid)
-			FSAL_UNSET_MASK(attrs->valid_mask, ATTR_GROUP);
-	}
+	/* For support_ex API, turn off owner and/or group attr
+	 * if they are the same as the credentials.
+	 */
+	if ((attrs->valid_mask & ATTR_OWNER) &&
+	    attrs->owner == op_ctx->creds->caller_uid)
+		FSAL_UNSET_MASK(attrs->valid_mask, ATTR_OWNER);
+
+	if ((attrs->valid_mask & ATTR_GROUP) &&
+	    attrs->group == op_ctx->creds->caller_gid)
+		FSAL_UNSET_MASK(attrs->valid_mask, ATTR_GROUP);
 
 	/* Permission checking will be done by the FSAL operation. */
 
@@ -854,18 +823,21 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 
 	switch (type) {
 	case REGULAR_FILE:
-		/* NOTE: will not be called here if support_ex */
-		status = parent->obj_ops.create(parent, name, attrs,
-						obj, attrs_out);
+		status = fsal_open2(parent, NULL, FSAL_O_RDWR, FSAL_UNCHECKED,
+				    name, attrs, NULL, obj, attrs_out);
+		if (FSAL_IS_SUCCESS(status)) {
+			/* Close it again; this is just a create */
+			(void)fsal_close(*obj);
+		}
 		break;
 
 	case DIRECTORY:
-		status = parent->obj_ops.mkdir(parent, name, attrs,
+		status = parent->obj_ops->mkdir(parent, name, attrs,
 					       obj, attrs_out);
 		break;
 
 	case SYMBOLIC_LINK:
-		status = parent->obj_ops.symlink(parent, name, link_content,
+		status = parent->obj_ops->symlink(parent, name, link_content,
 						 attrs, obj, attrs_out);
 		break;
 
@@ -873,7 +845,7 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 	case FIFO_FILE:
 	case BLOCK_FILE:
 	case CHARACTER_FILE:
-		status = parent->obj_ops.mknode(parent, name, type,
+		status = parent->obj_ops->mknode(parent, name, type,
 						attrs, obj, attrs_out);
 		break;
 
@@ -901,7 +873,7 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 					     "create failed because it already exists");
 				if ((*obj)->type != type) {
 					/* Incompatible types, returns NULL */
-					(*obj)->obj_ops.put_ref((*obj));
+					(*obj)->obj_ops->put_ref((*obj));
 					*obj = NULL;
 					goto out;
 				}
@@ -909,7 +881,7 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 				    (attrs->valid_mask & ATTR_SIZE) &&
 				    attrs->filesize == 0) {
 					attrs->valid_mask &= ATTR_SIZE;
-					goto setattrs;
+					goto out;
 				}
 			}
 		} else {
@@ -918,44 +890,6 @@ fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 		goto out;
 	}
 
-setattrs:
-	if (!support_ex) {
-		/* Handle setattr for old API */
-		attrs->valid_mask = orig_mask;
-		attrs->owner = owner;
-		attrs->group = group;
-
-		/* We already handled mode */
-		FSAL_UNSET_MASK(attrs->valid_mask, ATTR_MODE);
-
-		/* Check if attrs->owner was same as creds */
-		if ((attrs->valid_mask & ATTR_OWNER) &&
-		    (op_ctx->creds->caller_uid == attrs->owner))
-			FSAL_UNSET_MASK(attrs->valid_mask, ATTR_OWNER);
-
-		/* Check if attrs->group was same as creds */
-		if ((attrs->valid_mask & ATTR_GROUP) &&
-		    (op_ctx->creds->caller_gid == attrs->group))
-			FSAL_UNSET_MASK(attrs->valid_mask, ATTR_GROUP);
-
-		/* Setting uid/gid works only for root. AIX or Irix NFS
-                 * clients send gid on create if the parent directory has
-                 * setgid bit. Clear the OWNER and GROUP attributes for
-                 * create requests for non-root users.
-		 */
-		if (type != SYMBOLIC_LINK &&
-		    !op_ctx->fsal_export->exp_ops.is_superuser(
-					op_ctx->fsal_export, op_ctx->creds)) {
-			FSAL_UNSET_MASK(attrs->valid_mask,
-					ATTR_OWNER|ATTR_GROUP);
-		}
-
-		if (attrs->valid_mask) {
-			/* If any attributes were left to set, set them now. */
-			status = fsal_setattr(*obj, false, NULL, attrs);
-		}
-	}
-
  out:
 
 	/* Restore original mask so caller isn't bamboozled... */
@@ -989,7 +923,7 @@ bool fsal_create_verify(struct fsal_obj_handle *obj, uint32_t verf_hi,
 
 	fsal_prepare_attrs(&attrs, ATTR_ATIME | ATTR_MTIME);
 
-	obj->obj_ops.getattrs(obj, &attrs);
+	obj->obj_ops->getattrs(obj, &attrs);
 	if (FSAL_TEST_MASK(attrs.valid_mask, ATTR_ATIME)
 	    && FSAL_TEST_MASK(attrs.valid_mask, ATTR_MTIME)
 	    && attrs.atime.tv_sec == verf_hi
@@ -1002,283 +936,6 @@ bool fsal_create_verify(struct fsal_obj_handle *obj, uint32_t verf_hi,
 	return verified;
 }
 
-/**
- * @brief New style reads
- *
- * @param[in]     obj          File to be read or written
- * @param[in]     bypass       If state doesn't indicate a share reservation,
- *                             bypass any deny read
- * @param[in]     state        state_t associated with the operation
- * @param[in]     offset       Absolute file position for I/O
- * @param[in]     io_size      Amount of data to be read or written
- * @param[out]    bytes_moved  The length of data successfuly read
- * @param[in,out] buffer       Where in memory to read data
- * @param[out]    eof          Whether a READ encountered the end of file
- * @param[in]     info         io_info for READ_PLUS
- *
- * @return FSAL status
- */
-
-fsal_status_t fsal_read2(struct fsal_obj_handle *obj,
-			 bool bypass,
-			 struct state_t *state,
-			 uint64_t offset,
-			 size_t io_size,
-			 size_t *bytes_moved,
-			 void *buffer,
-			 bool *eof,
-			 struct io_info *info)
-{
-	/* Error return from FSAL calls */
-	fsal_status_t status = { 0, 0 };
-
-	status = obj->obj_ops.read2(obj, bypass, state, offset, io_size, buffer,
-				    bytes_moved, eof, info);
-
-	/* Fixup FSAL_SHARE_DENIED status */
-	if (status.major == ERR_FSAL_SHARE_DENIED)
-		status = fsalstat(ERR_FSAL_LOCKED, 0);
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "FSAL READ operation returned %s, asked_size=%zu, effective_size=%zu",
-		     fsal_err_txt(status), io_size, *bytes_moved);
-
-	if (FSAL_IS_ERROR(status)) {
-		*bytes_moved = 0;
-		return status;
-	}
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "inode/direct: io_size=%zu, bytes_moved=%zu, offset=%"
-		     PRIu64, io_size, *bytes_moved, offset);
-
-	return fsalstat(ERR_FSAL_NO_ERROR, 0);
-}
-
-/**
- * @brief New style writes
- *
- * @param[in]     obj          File to be read or written
- * @param[in]     bypass       If state doesn't indicate a share reservation,
- *                             bypass any non-mandatory deny write
- * @param[in]     state        state_t associated with the operation
- * @param[in]     offset       Absolute file position for I/O
- * @param[in]     io_size      Amount of data to be written
- * @param[out]    bytes_moved  The length of data successfuly written
- * @param[in,out] buffer       Where in memory to write data
- * @param[in]     sync         Whether the write is synchronous or not
- * @param[in]     info         io_info for WRITE_PLUS
- *
- * @return FSAL status
- */
-
-fsal_status_t fsal_write2(struct fsal_obj_handle *obj,
-			  bool bypass,
-			  struct state_t *state,
-			  uint64_t offset,
-			  size_t io_size,
-			  size_t *bytes_moved,
-			  void *buffer,
-			  bool *sync,
-			  struct io_info *info)
-{
-	/* Error return from FSAL calls */
-	fsal_status_t status = { 0, 0 };
-
-	if (op_ctx->export_perms->options & EXPORT_OPTION_COMMIT) {
-		/* Force sync if export requires it */
-		*sync = true;
-	}
-
-	status = obj->obj_ops.write2(obj,
-				     bypass,
-				     state,
-				     offset,
-				     io_size,
-				     buffer,
-				     bytes_moved,
-				     sync,
-				     info);
-
-	/* Fixup ERR_FSAL_SHARE_DENIED status */
-	if (status.major == ERR_FSAL_SHARE_DENIED)
-		status = fsalstat(ERR_FSAL_LOCKED, 0);
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "FSAL WRITE operation returned %s, asked_size=%zu, effective_size=%zu",
-		     fsal_err_txt(status), io_size, *bytes_moved);
-
-	if (FSAL_IS_ERROR(status)) {
-		*bytes_moved = 0;
-
-		return status;
-	}
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "inode/direct: io_size=%zu, bytes_moved=%zu, offset=%"
-		     PRIu64, io_size, *bytes_moved, offset);
-
-	return fsalstat(ERR_FSAL_NO_ERROR, 0);
-}
-
-/**
- * @brief Read/Write
- *
- * @param[in]     obj          File to be read or written
- * @param[in]     io_direction Whether this is a read or a write
- * @param[in]     offset       Absolute file position for I/O
- * @param[in]     io_size      Amount of data to be read or written
- * @param[out]    bytes_moved  The length of data successfuly read or written
- * @param[in,out] buffer       Where in memory to read or write data
- * @param[out]    eof          Whether a READ encountered the end of file.  May
- *                             be NULL for writes.
- * @param[in]     sync         Whether the write is synchronous or not
- *
- * @return FSAL status
- */
-
-fsal_status_t fsal_rdwr(struct fsal_obj_handle *obj,
-		      fsal_io_direction_t io_direction,
-		      uint64_t offset, size_t io_size,
-		      size_t *bytes_moved, void *buffer,
-		      bool *eof,
-		      bool *sync, struct io_info *info)
-{
-	/* Error return from FSAL calls */
-	fsal_status_t fsal_status = { 0, 0 };
-	/* Required open mode to successfully read or write */
-	fsal_openflags_t openflags = FSAL_O_CLOSED;
-	fsal_openflags_t loflags;
-	/* TRUE if we opened a previously closed FD */
-	bool opened = false;
-
-	/* Set flags for a read or write, as appropriate */
-	if (io_direction == FSAL_IO_READ ||
-	    io_direction == FSAL_IO_READ_PLUS) {
-		openflags = FSAL_O_READ;
-	} else {
-		/* Pretent that the caller requested sync (stable write)
-		 * if the export has COMMIT option.
-		 */
-		if (op_ctx->export_perms->options & EXPORT_OPTION_COMMIT)
-			*sync = true;
-		openflags = FSAL_O_WRITE;
-	}
-
-	assert(obj != NULL);
-
-	/* IO is done only on REGULAR_FILEs */
-	if (obj->type != REGULAR_FILE) {
-		fsal_status = fsalstat(
-		    obj->type ==
-		    DIRECTORY ? ERR_FSAL_ISDIR :
-		    ERR_FSAL_BADTYPE, 0);
-		goto out;
-	}
-
-	loflags = obj->obj_ops.status(obj);
-	while ((!fsal_is_open(obj))
-	       || (loflags && loflags != FSAL_O_RDWR && loflags != openflags)) {
-		loflags = obj->obj_ops.status(obj);
-		if ((!fsal_is_open(obj))
-		    || (loflags && loflags != FSAL_O_RDWR
-			&& loflags != openflags)) {
-			fsal_status = fsal_open(obj, openflags);
-			if (FSAL_IS_ERROR(fsal_status))
-				goto out;
-			opened = true;
-		}
-		loflags = obj->obj_ops.status(obj);
-	}
-
-	/* Call FSAL_read or FSAL_write */
-	if (io_direction == FSAL_IO_READ) {
-		fsal_status = obj->obj_ops.read(obj, offset, io_size, buffer,
-						bytes_moved, eof);
-	} else if (io_direction == FSAL_IO_READ_PLUS) {
-		fsal_status = obj->obj_ops.read_plus(obj, offset, io_size,
-						     buffer, bytes_moved, eof,
-						     info);
-	} else {
-		bool fsal_sync = *sync;
-
-		if (io_direction == FSAL_IO_WRITE)
-			fsal_status = obj->obj_ops.write(obj, offset, io_size,
-							 buffer, bytes_moved,
-							 &fsal_sync);
-		else
-			fsal_status = obj->obj_ops.write_plus(obj, offset,
-							      io_size, buffer,
-							      bytes_moved,
-							      &fsal_sync, info);
-		/* Alright, the unstable write is complete. Now if it was
-		   supposed to be a stable write we can sync to the hard
-		   drive. */
-
-		if (*sync && !fsal_sync && !FSAL_IS_ERROR(fsal_status)) {
-			fsal_status = obj->obj_ops.commit(obj, offset, io_size);
-		} else {
-			*sync = fsal_sync;
-		}
-	}
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "fsal_rdwr_plus: FSAL IO operation returned %s, asked_size=%zu, effective_size=%zu",
-		     fsal_err_txt(fsal_status), io_size, *bytes_moved);
-
-	if (FSAL_IS_ERROR(fsal_status)) {
-		if (fsal_status.major == ERR_FSAL_DELAY) {
-			LogEvent(COMPONENT_FSAL,
-				 "fsal_rdwr_plus: FSAL_write  returned EBUSY");
-		} else {
-			LogDebug(COMPONENT_FSAL,
-				 "fsal_rdwr_plus: fsal_status = %s",
-				 fsal_err_txt(fsal_status));
-		}
-
-		*bytes_moved = 0;
-
-		if (fsal_status.major == ERR_FSAL_STALE)
-			goto out;
-
-		if ((fsal_status.major != ERR_FSAL_NOT_OPENED)
-		    && (obj->obj_ops.status(obj) != FSAL_O_CLOSED)) {
-			LogFullDebug(COMPONENT_FSAL,
-				     "fsal_rdwr_plus: CLOSING file %p",
-				     obj);
-
-			fsal_status = obj->obj_ops.close(obj);
-			if (FSAL_IS_ERROR(fsal_status)) {
-				LogCrit(COMPONENT_FSAL,
-					"Error closing file in fsal_rdwr_plus: %s.",
-					fsal_err_txt(fsal_status));
-			}
-		}
-
-		goto out;
-	}
-
-	LogFullDebug(COMPONENT_FSAL,
-		     "fsal_rdwr_plus: inode/direct: io_size=%zu, bytes_moved=%zu, offset=%"
-		     PRIu64, io_size, *bytes_moved, offset);
-
-	if (opened) {
-		fsal_status = obj->obj_ops.close(obj);
-		if (FSAL_IS_ERROR(fsal_status)) {
-			LogEvent(COMPONENT_FSAL,
-				 "fsal_rdwr_plus: close = %s",
-				 fsal_err_txt(fsal_status));
-			goto out;
-		}
-	}
-
-	fsal_status = fsalstat(0, 0);
-
- out:
-
-	return fsal_status;
-}
-
 struct fsal_populate_cb_state {
 	struct fsal_obj_handle *directory;
 	fsal_status_t *status;
@@ -1344,6 +1001,7 @@ populate_dirent(const char *name,
 						 0, cookie, state->cb_state);
 				/* Protocol layers NEVER do readahead. */
 				retval = DIR_TERMINATE;
+				put_gsh_export(junction_export);
 				goto out;
 			}
 		} else {
@@ -1367,7 +1025,7 @@ populate_dirent(const char *name,
 					.fs_supported_attrs(op_ctx->fsal_export)
 					| ATTR_RDATTR_ERR);
 
-		status = junction_obj->obj_ops.getattrs(junction_obj, &attrs2);
+		status = junction_obj->obj_ops->getattrs(junction_obj, &attrs2);
 
 		if (!FSAL_IS_ERROR(status)) {
 			/* Now call the callback again with that. */
@@ -1388,7 +1046,7 @@ populate_dirent(const char *name,
 		/* Release our refs */
 		op_ctx->fsal_export = saved_export;
 
-		junction_obj->obj_ops.put_ref(junction_obj);
+		junction_obj->obj_ops->put_ref(junction_obj);
 		put_gsh_export(junction_export);
 	}
 
@@ -1403,7 +1061,7 @@ populate_dirent(const char *name,
 out:
 
 	/* Put the ref on obj that readdir took */
-	obj->obj_ops.put_ref(obj);
+	obj->obj_ops->put_ref(obj);
 
 	return retval;
 }
@@ -1453,8 +1111,7 @@ fsal_status_t fsal_readdir(struct fsal_obj_handle *directory,
 
 	/* readdir can be done only with a directory */
 	if (directory->type != DIRECTORY) {
-		LogFullDebug(COMPONENT_NFS_READDIR,
-			     "Not a directory");
+		LogDebug(COMPONENT_NFS_READDIR, "Not a directory");
 		return fsalstat(ERR_FSAL_NOTDIR, 0);
 	}
 
@@ -1468,9 +1125,9 @@ fsal_status_t fsal_readdir(struct fsal_obj_handle *directory,
 
 	fsal_status = fsal_access(directory, access_mask);
 	if (FSAL_IS_ERROR(fsal_status)) {
-		LogFullDebug(COMPONENT_NFS_READDIR,
-			     "permission check for directory status=%s",
-			     fsal_err_txt(fsal_status));
+		LogDebug(COMPONENT_NFS_READDIR,
+			 "permission check for directory status=%s",
+			 fsal_err_txt(fsal_status));
 		return fsal_status;
 	}
 	if (attrmask != 0) {
@@ -1478,9 +1135,9 @@ fsal_status_t fsal_readdir(struct fsal_obj_handle *directory,
 		fsal_status_t attr_status = fsal_access(directory,
 							access_mask_attr);
 		if (FSAL_IS_ERROR(attr_status))
-			LogFullDebug(COMPONENT_NFS_READDIR,
-				     "permission check for attributes status=%s",
-				     fsal_err_txt(attr_status));
+			LogDebug(COMPONENT_NFS_READDIR,
+				 "permission check for attributes status=%s",
+				 fsal_err_txt(attr_status));
 		state.cb_parms.attr_allowed = !FSAL_IS_ERROR(attr_status);
 	} else {
 		/* No attributes requested. */
@@ -1498,7 +1155,7 @@ fsal_status_t fsal_readdir(struct fsal_obj_handle *directory,
 	state.cb_nfound = nbfound;
 	state.attrmask = attrmask;
 
-	fsal_status = directory->obj_ops.readdir(directory, &cookie,
+	fsal_status = directory->obj_ops->readdir(directory, &cookie,
 						 (void *)&state,
 						 populate_dirent,
 						 attrmask,
@@ -1564,7 +1221,7 @@ fsal_remove(struct fsal_obj_handle *parent, const char *name)
 		goto out;
 #endif /* ENABLE_RFC_ACL */
 
-	status = parent->obj_ops.unlink(parent, to_remove_obj, name);
+	status = parent->obj_ops->unlink(parent, to_remove_obj, name);
 
 	if (FSAL_IS_ERROR(status)) {
 		LogFullDebug(COMPONENT_FSAL, "unlink %s failure %s",
@@ -1574,7 +1231,7 @@ fsal_remove(struct fsal_obj_handle *parent, const char *name)
 
 out:
 
-	to_remove_obj->obj_ops.put_ref(to_remove_obj);
+	to_remove_obj->obj_ops->put_ref(to_remove_obj);
 
 out_no_obj:
 
@@ -1606,9 +1263,10 @@ fsal_status_t fsal_rename(struct fsal_obj_handle *dir_src,
 		return fsalstat(ERR_FSAL_NOTDIR, 0);
 
 	/* Check for . and .. on oldname and newname. */
-	if (!strcmp(oldname, ".") || !strcmp(oldname, "..")
+	if (oldname[0] == '\0' || newname[0] == '\0'
+	    || !strcmp(oldname, ".") || !strcmp(oldname, "..")
 	    || !strcmp(newname, ".") || !strcmp(newname, "..")) {
-		return fsalstat(ERR_FSAL_BADNAME, 0);
+		return fsalstat(ERR_FSAL_INVAL, 0);
 	}
 
 	/* Check for object existence in source directory */
@@ -1628,9 +1286,15 @@ fsal_status_t fsal_rename(struct fsal_obj_handle *dir_src,
 		goto out;
 	}
 
+	/* Don't allow rename of an object as parent of itself */
+	if (dir_dest == lookup_src) {
+		fsal_status = fsalstat(ERR_FSAL_INVAL, 0);
+		goto out;
+	}
+
 	LogFullDebug(COMPONENT_FSAL, "about to call FSAL rename");
 
-	fsal_status = dir_src->obj_ops.rename(lookup_src, dir_src, oldname,
+	fsal_status = dir_src->obj_ops->rename(lookup_src, dir_src, oldname,
 					      dir_dest, newname);
 
 	LogFullDebug(COMPONENT_FSAL, "returned from FSAL rename");
@@ -1650,76 +1314,12 @@ out:
 		 * export since that would be the junction node, NOT the export
 		 * root node on the other side of the junction.
 		 */
-		lookup_src->obj_ops.put_ref(lookup_src);
+		lookup_src->obj_ops->put_ref(lookup_src);
 	}
 
 	return fsal_status;
 }
 
-/**
- * @brief Open a file
- *
- * @param[in] obj	File to open
- * @param[in] openflags	The type of access for which to open
- * @return FSAL status
- */
-fsal_status_t fsal_open(struct fsal_obj_handle *obj_hdl,
-			fsal_openflags_t openflags)
-{
-	fsal_openflags_t current_flags;
-	fsal_status_t status = {0, 0};
-
-	if (obj_hdl->type != REGULAR_FILE)
-		return fsalstat(ERR_FSAL_BADTYPE, 0);
-
-	current_flags = obj_hdl->obj_ops.status(obj_hdl);
-
-	/* Filter out overloaded FSAL_O_RECLAIM */
-	openflags &= ~FSAL_O_RECLAIM;
-
-	/* Make sure current state meet requirements */
-	if ((current_flags != FSAL_O_RDWR) && (current_flags != FSAL_O_CLOSED)
-	    && (current_flags != openflags)) {
-		bool closed;
-		/* Flags are insufficient; need to re-open */
-		if (op_ctx->fsal_export->exp_ops.fs_supports(
-			op_ctx->fsal_export, fso_reopen_method)) {
-			/* FSAL has re-open; use that */
-			status = obj_hdl->obj_ops.reopen(obj_hdl,
-							   openflags);
-			closed = false;
-		} else {
-			status = obj_hdl->obj_ops.close(obj_hdl);
-			closed = true;
-		}
-		if (FSAL_IS_ERROR(status)
-		    && (status.major != ERR_FSAL_NOT_OPENED))
-			return status;
-		if (!FSAL_IS_ERROR(status) && closed) {
-			(void) atomic_dec_size_t(&open_fd_count);
-		}
-
-		/* Potentially force re-openning */
-		current_flags = obj_hdl->obj_ops.status(obj_hdl);
-	}
-
-	if (current_flags == FSAL_O_CLOSED) {
-		status = obj_hdl->obj_ops.open(obj_hdl, openflags);
-		if (FSAL_IS_ERROR(status))
-			return status;
-
-		(void) atomic_inc_size_t(&open_fd_count);
-
-		LogDebug(COMPONENT_FSAL,
-			 "obj %p: openflags = %d, open_fd_count = %zd",
-			 obj_hdl, openflags,
-			 atomic_fetch_size_t(&open_fd_count));
-	}
-	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
-
-	return status;
-}
-
 /**
  * @brief Opens a file by name or by handle.
  *
@@ -1816,7 +1416,7 @@ fsal_status_t fsal_open2(struct fsal_obj_handle *in_obj,
 	 * might be an exclusive recreate replay and we want the FSAL to
 	 * check the verifier.
 	 */
-	status = in_obj->obj_ops.open2(in_obj,
+	status = in_obj->obj_ops->open2(in_obj,
 				       state,
 				       openflags,
 				       createmode,
@@ -1830,7 +1430,7 @@ fsal_status_t fsal_open2(struct fsal_obj_handle *in_obj,
 	if (!FSAL_IS_ERROR(status)) {
 		/* Get a reference to the entry. */
 		*obj = in_obj;
-		in_obj->obj_ops.get_ref(in_obj);
+		in_obj->obj_ops->get_ref(in_obj);
 	}
 
 	return status;
@@ -1867,7 +1467,7 @@ fsal_status_t fsal_reopen2(struct fsal_obj_handle *obj,
 
 	/* Re-open the entry in the FSAL.
 	 */
-	status = obj->obj_ops.reopen2(obj, state, openflags);
+	status = obj->obj_ops->reopen2(obj, state, openflags);
 
  out:
 
@@ -1892,7 +1492,7 @@ fsal_status_t fsal_statfs(struct fsal_obj_handle *obj,
 	fsal_status =
 	    export->exp_ops.get_fs_dynamic_info(export, obj, dynamicinfo);
 	LogFullDebug(COMPONENT_FSAL,
-		     "fsal_statfs: dynamicinfo: {total_bytes = %" PRIu64
+		     "dynamicinfo: {total_bytes = %" PRIu64
 		     ", free_bytes = %" PRIu64 ", avail_bytes = %" PRIu64
 		     ", total_files = %" PRIu64 ", free_files = %" PRIu64
 		     ", avail_files = %" PRIu64 "}", dynamicinfo->total_bytes,
@@ -1902,43 +1502,6 @@ fsal_status_t fsal_statfs(struct fsal_obj_handle *obj,
 	return fsal_status;
 }
 
-/**
- * @brief Commit a section of a file to storage
- *
- * @param[in] obj	File to commit
- * @param[in] offset	Offset for start of commit
- * @param[in] len	Length of commit
- * @return FSAL status
- */
-fsal_status_t fsal_commit(struct fsal_obj_handle *obj, off_t offset,
-			 size_t len)
-{
-	/* Error return from FSAL calls */
-	fsal_status_t fsal_status = { 0, 0 };
-	bool opened = false;
-
-	if ((uint64_t) len > ~(uint64_t) offset)
-		return fsalstat(ERR_FSAL_INVAL, 0);
-
-	if (obj->fsal->m_ops.support_ex(obj))
-		return obj->obj_ops.commit2(obj, offset, len);
-
-	if (!fsal_is_open(obj)) {
-		LogFullDebug(COMPONENT_FSAL, "need to open");
-		fsal_status = fsal_open(obj, FSAL_O_WRITE);
-		if (FSAL_IS_ERROR(fsal_status))
-			return fsal_status;
-		opened = true;
-	}
-
-	fsal_status = obj->obj_ops.commit(obj, offset, len);
-
-	if (opened)
-		obj->obj_ops.close(obj);
-
-	return fsal_status;
-}
-
 /**
  * @brief Verify an exclusive create replay when the file is already open.
  *
@@ -1954,7 +1517,7 @@ fsal_status_t fsal_commit(struct fsal_obj_handle *obj, off_t offset,
 fsal_status_t fsal_verify2(struct fsal_obj_handle *obj,
 			   fsal_verifier_t verifier)
 {
-	if (!obj->obj_ops.check_verifier(obj, verifier)) {
+	if (!obj->obj_ops->check_verifier(obj, verifier)) {
 		/* Verifier check failed. */
 		return fsalstat(ERR_FSAL_EXIST, 0);
 	}
@@ -1962,4 +1525,42 @@ fsal_status_t fsal_verify2(struct fsal_obj_handle *obj,
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
+/**
+ * @brief Fetch optional attributes
+ *
+ * The request_mask should be set in attrs_out indicating which attributes
+ * are desired. If ATTR_RDATTR_ERR is set, and the getattrs fails,
+ * the error ERR_FSAL_NO_ERROR will be returned, however the attributes
+ * valid_mask will be set to ATTR_RDATTR_ERR. Otherwise, if
+ * ATTR_RDATTR_ERR is not set and the getattrs fails, the error returned
+ * by getattrs will be returned.
+ *
+ * @param[in]     obj_hdl   Object to get attributes for.
+ * @param[in,out] attrs_out Optional attributes for the object
+ *
+ * @return FSAL status.
+ **/
+fsal_status_t get_optional_attrs(struct fsal_obj_handle *obj_hdl,
+				 struct attrlist *attrs_out)
+{
+	fsal_status_t status;
+
+	if (attrs_out == NULL)
+		return fsalstat(ERR_FSAL_NO_ERROR, 0);
+
+	status = obj_hdl->obj_ops->getattrs(obj_hdl, attrs_out);
+
+	if (FSAL_IS_ERROR(status)) {
+		if (attrs_out->request_mask & ATTR_RDATTR_ERR) {
+			/* Indicate the failure of requesting attributes by
+			 * marking the ATTR_RDATTR_ERR in the mask.
+			 */
+			attrs_out->valid_mask = ATTR_RDATTR_ERR;
+			status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+		} /* otherwise let the error stand. */
+	}
+
+	return status;
+}
+
 /** @} */
